# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
import yaml
import sys
import os
import tempfile
import traceback
import subprocess
import array
import cairo
import ftplib
import random
import string
import datetime

# import math
try:
    from gi import require_version
except ImportError:
    print("ERROR: pygobject is not installed.")
    raise
try:
    require_version("Gtk", "3.0")
except ValueError:
    print("ERROR: gtk3 development files are not installed.")
    raise
except AttributeError:
    print("pygobject version too old.")
    raise
try:
    require_version('PangoCairo', '1.0')
except ValueError:
    print("ERROR: PangoCairo development files are not installed.")
    raise
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Pango
from gi.repository import PangoCairo
from os.path import abspath
from os.path import dirname
from os.path import isfile
from os.path import join
from pkg_resources import resource_exists
from pkg_resources import resource_filename
from logging import getLogger
from logging import basicConfig
from logging import INFO
from six import iteritems
from six import StringIO
# from .labelprintingop import PrintingApp
from .jscript_utils import shift_and_format_items
from .pw_utils import PasswordGenerator
try:
    from synapselabels import __version__
    assert __version__
except ImportError:
    __version__ = "UNKNOWN"


log = getLogger('synapselabels')
dnd_targets = [('text/plain', Gtk.TargetFlags.OTHER_WIDGET, 0), ]
default_unit = 'm'


def hex_increment(hex_string):
    last = 1 + int(hex_string[-1:], 16)
    if last <= 15:
        return hex_string[:-1].upper() + ('%X' % last)
    else:
        return hex_increment(hex_string[:-1]) + '0'


def dec_increment(dec_string):
    last = 1 + int(dec_string[-1:])
    if last <= 9:
        return dec_string[:-1] + str(last)
    else:
        return dec_increment(dec_string[:-1]) + '0'


def start_file(filepath):
    if sys.platform.startswith('darwin'):
        subprocess.call(('open', filepath))
    elif os.name == 'nt':
        os.startfile(filepath)
    elif os.name == 'posix':
        subprocess.call(('xdg-open', filepath))


def get_path(name):
    """
    Returns the absolute path name of a resource file
    File in the current directory overrides file in package
    """
    path = None
    if isfile(name):
        path = abspath(name)
    elif resource_exists(__name__, name):
        res_filename = resource_filename(__name__, name)
        res_path = abspath(join(dirname(__file__), res_filename))
        if isfile(res_path):
            path = res_path
    if path:
        log.info("located path: %s", path)
    else:
        log.error("could not locate path for: %s", name)
    return path


label_config_file = get_path('labelconfig.yaml')
glade_file = get_path('labelsgui.glade')
src1 = get_path('names.txt')
src2 = get_path('places.txt')


def check_parameter(parameter, value):
    """raises ValueError if parameter does not comply"""
    if parameter == 'date code':
        if len(value) != 4:
            raise ValueError('date code must be 4 characters long: YYWW')
        try:
            year = int(value[:2])
        except ValueError:
            raise ValueError('YY in date code (YYWW) must be digits')
        if year < 7:
            raise ValueError(
                'YY in date code (YYWW) must not represent a year'
                ' less than 2007, when Synapse was founded'
                )
        try:
            week = int(value[2:])
        except ValueError:
            raise ValueError('WW in date code (YYWW) must be digits')
        if week > 52:
            raise ValueError('WW in date code (YYWW) is too large')
    elif parameter == 'ethernet mac address':
        if len(value) != 12:
            raise ValueError('ethernet mac address must be 12 characters long')
        if value[:6] != '001C2C':
            raise ValueError('ethernet mac address must start with 001C2C')
        try:
            int(value[6:], 16)
        except ValueError:
            raise ValueError('ethernet mac address must consist of hex digits')
    elif parameter == 'ethernet mac address last 6':
        if len(value) != 6:
            raise ValueError(
                'ethernet mac address last 6 must be 6 characters long'
                )
        try:
            int(value, 16)
        except ValueError:
            raise ValueError(
                'ethernet mac address last 6 must consist of hex digits'
                )
    elif parameter == 'wifi mac address':
        if len(value) != 12:
            raise ValueError(
                'wifi mac address must be 12 characters long'
                )
        try:
            int(value, 16)
        except ValueError:
            raise ValueError(
                'wifi mac address must consist of hex digits'
                )
    elif parameter == 'snap mac address':
        if len(value) != 16:
            raise ValueError('snap mac address must be 16 characters long')
        if value[:6] != '001C2C':
            raise ValueError('snap mac address must start with 001C2C')
        try:
            int(value[6:], 16)
        except ValueError:
            raise ValueError('snap mac address must consist of hex digits')
    elif parameter == 'snap node address':
        if len(value) != 6:
            raise ValueError(
                'snap node address must be 6 characters long'
                )
        try:
            int(value, 16)
        except ValueError:
            raise ValueError(
                'snap node address must consist of hex digits'
                )
    elif parameter == 'modem meid':
        if len(value) != 14:
            raise ValueError(
                'modem meid must be 14 characters long'
                )
        try:
            int(value, 16)
        except ValueError:
            raise ValueError(
                'modem meid must consist of hex digits'
                )
    elif parameter == 'sim iccid':
        if len(value) != 20:
            raise ValueError(
                'sim iccid must be 20 characters long'
                )
        try:
            int(value)
        except ValueError:
            raise ValueError(
                'sim iccid must consist of digits'
                )


def gtk_input(
    parent,
    message,
    default='',
    auto_inc_option=False,
        fixed_option=False):
    """
    Display a dialog with a text entry.
    Returns the text, or None if canceled.
    """
    if auto_inc_option and fixed_option:
        raise ValueError('Both Auto-increment and Fixed')
    d = Gtk.MessageDialog(
        parent,
        Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
        Gtk.MessageType.QUESTION,
        Gtk.ButtonsType.OK_CANCEL,
        message
        )
    entry = Gtk.Entry()
    entry.set_text(default)
    entry.show()
    if auto_inc_option:
        auto_inc = Gtk.CheckButton(label='Auto-increment')
        auto_inc.show()
        d.vbox.pack_end(auto_inc, True, True, 0)
    if fixed_option:
        fixed = Gtk.CheckButton(label='Fixed')
        fixed.show()
        d.vbox.pack_end(fixed, True, True, 0)
    d.vbox.pack_end(entry, True, True, 0)
    entry.connect('activate', lambda _: d.response(Gtk.ResponseType.OK))
    d.set_default_response(Gtk.ResponseType.OK)

    r = d.run()
    text = entry.get_text().decode('utf8')
    d.destroy()
    if r == Gtk.ResponseType.OK:
        return (
            text,
            auto_inc.get_active() if auto_inc_option else
            fixed.get_active() if fixed_option else
            False
            )


def dots(size, unit, dpi):
    if size.startswith('pt'):
        size = float(size[2:],) / 72.0
    else:
        size = float(size)
        if unit == 'm':
            size = size / 25.4  # convert mm to inch
        elif unit != 'i':
            raise ValueError(
                'invalid unit: %s' % repr(unit)
                )
    return dpi * size


def pil2cairo(im):
    """Transform a PIL Image into a Cairo ImageSurface."""

    assert sys.byteorder == 'little', 'We don\'t support big endian'
    if im.mode != 'RGBA':
        im = im.convert('RGBA')

    s = im.tobytes('raw', 'BGRA')
    a = array.array('B', s)
    dest = cairo.ImageSurface(cairo.FORMAT_ARGB32, im.size[0], im.size[1])
    ctx = cairo.Context(dest)
    non_premult_src_wo_alpha = cairo.ImageSurface.create_for_data(
        a, cairo.FORMAT_RGB24, im.size[0], im.size[1])
    non_premult_src_alpha = cairo.ImageSurface.create_for_data(
        a, cairo.FORMAT_ARGB32, im.size[0], im.size[1])
    ctx.set_source_surface(non_premult_src_wo_alpha)
    ctx.mask_surface(non_premult_src_alpha)
    return dest


def get_selected_data(treeview):
        # print("print_selected_label")
        (model, pathlist) = treeview.get_selection().get_selected_rows()
        for path in pathlist:
            tree_iter = model.get_iter(path)
            for index, column in enumerate(
                    treeview.get_columns()):
                return model.get_value(tree_iter, index)


def on_drag_data_get(treeview, context, selection, info, timestamp):
    """ Extract data from the source of the DnD operation.

    Serialize iterators of selected tasks in format
    <iter>,<iter>,...,<iter> and set it as parameter of DND """

    selected_data = get_selected_data(treeview)
    selection.set(selection.get_target(), 0, selected_data)


def set_treeview_columns(treeview, columns):
    for i in range(len(columns)):
        # cellrenderer to render the text
        cell = Gtk.CellRendererText(editable=False)
        # cell.set_property("editable", True)
        # the text in the first column should be in boldface
        # if i == 0:
        #    cell.props.weight_set = True
        #    cell.props.weight = Pango.Weight.BOLD
        # the column is created
        col = Gtk.TreeViewColumn(
            columns[i], cell, text=i)
        # and it is appended to the treeview
        treeview.append_column(col)


def set_treeview_from_clipboard(treeview, with_header=True):
    set_treeview_from_text_data(
        treeview,
        Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).wait_for_text(),
        with_header
        )


def set_treeview_from_text_data(treeview, text_data, with_header=True):
    listmodel = None
    if text_data is not None:
        # print("clipboard text:", repr(text_data))
        for line in text_data.splitlines():
            # print("line:", repr(line))
            texts = line.split('\t')
            if listmodel is None:
                for column in treeview.get_columns():
                    treeview.remove_column(column)
                args = [GObject.TYPE_STRING] * len(texts)
                # print(args)
                listmodel = Gtk.ListStore(*args)
                # listmodel = Gtk.ListStore(str, str, str)
                if with_header:
                    columns = texts
                else:
                    columns = [
                        "Column{}".format(index) for
                        index in
                        range(0, len(texts))
                        ]
                    # print("appending:", texts)
                    listmodel.append(texts)
            else:
                # print("appending:", texts)
                listmodel.append(texts)
        if listmodel is not None:
            treeview.set_model(listmodel)
            set_treeview_columns(treeview, columns)
    else:
        print("No text on the clipboard.")


def set_treeview_from_texts(treeview, texts):
    # treeview.set_headers_visible(False)
    listmodel = None
    if texts is not None:
        # print("texts:", repr(texts))
        for line in texts:
            # print("line:", repr(line))
            if listmodel is None:
                for column in treeview.get_columns():
                    treeview.remove_column(column)
                listmodel = Gtk.ListStore(GObject.TYPE_STRING)
                # print("appending:", line)
                listmodel.append([line])
            else:
                # print("appending:", line)
                listmodel.append([line])
        if listmodel is not None:
            # print("setting model:", listmodel)
            treeview.set_model(listmodel)
            set_treeview_columns(treeview, ['Params'])


def fill_combo(combo, data):
    store = Gtk.ListStore(GObject.TYPE_STRING)
    for line in data:
        store.append([line])
    combo.set_model(store)
    combo.set_active(0)
    # And here's the new stuff:
    cell = Gtk.CellRendererText()
    combo.pack_start(cell, True)
    combo.add_attribute(cell, "text", 0)


def get_treeview_column(treeview, x_pos):
    for column in treeview.get_columns():
        if column.get_x_offset() <= x_pos:
            if x_pos <= (column.get_x_offset() + column.get_width()):
                return column


def drag_data_received(
        widget, context, x_pos, y_pos, selection, info, timestamp):
    data = selection.get_data()
    print('drag_data_received data:', data)
    if data:
        column = get_treeview_column(widget, x_pos)
        if column:
            print(
                "old column title:", repr(column.get_title()),
                "new column title:", repr(data)
                )
            column.set_title(data)


class LabelConfig(object):

    def __init__(self, config_file_path):
        with open(config_file_path, 'r') as stream:
            yaml_data = yaml.load(stream)
        # print('yaml_data:', repr(yaml_data))
        # yaml.dump(yaml_data, sys.stdout, default_flow_style=False)
        if 'label types' in yaml_data:
            self.label_types = yaml_data['label types']
            # print('label types:', repr(self.label_types))
        else:
            raise ValueError('"label types" is not found in config file')
        if 'printers' in yaml_data:
            self.printers = yaml_data['printers']
            # print('printers:', repr(printers))
        else:
            raise ValueError('"printers" is not found in config file')
        if 'models' in yaml_data:
            self.models = yaml_data['models']
            # print('models:', repr(self.models))
            self.model_names = []
            for model in self.models:
                for model_name in model:
                    self.model_names.append(model_name)
        else:
            raise ValueError('"models" is not found in config file')
        if 'labels' in yaml_data:
            self.labels = yaml_data['labels']
        else:
            raise ValueError('"labels" is not found in config file')


class LabelsGui(LabelConfig):
    """
    This is the E20 GUI
    """
    HEADER_HEIGHT = 10 * 72 / 25.4
    HEADER_GAP = 3 * 72 / 25.4

    def __init__(self):
        """
        Build GUI
        """
        super(LabelsGui, self).__init__(label_config_file)
        self.incremented_params = {}
        self.fixed_params = {}
        self.model = None
        self.dpi = 300
        # Build GUI from Glade file
        self.builder = Gtk.Builder()
        self.builder.add_from_file(glade_file)
        # Get objects
        go = self.builder.get_object
        self.window = go('applicationwindow1')
        self.window.set_title('Synapse Labels - v{}'.format(__version__))
        self.textview = go('textview')
        self.drawingarea = go('drawingarea')
        self.drawingarea.connect("draw", self.on_draw)
        self.treeview_data_table = go('treeviewDataTable')
        self.treeview_data_table.get_selection().connect(
            "changed",
            self.data_tree_selection_changed,
            "dummy"
            )
        self.treeview_data_table.enable_model_drag_dest(
            dnd_targets,
            Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE
            )
        self.treeview_data_table.connect(
            'drag-data-received',
            drag_data_received,
            )
        # self.combobox = go('comboboxtext1')
        # self.mac_num = go('labelMacNum')
        self.entry_date_code = go('entryDateCode')
        self.entry_records = go('entryRecords')
        self.entry_records.connect(
            "changed",
            self.entry_records_changed
            )
        self.button_today = go('buttonToday')
        self.entry_copies = go('entryCopies')
        self.button_today.connect(
            "clicked",
            self.button_today_clicked,
            "today button data"
            )
        self.button_generate = go('buttonGenerate')
        self.button_generate.connect(
            "clicked",
            self.button_generate_clicked,
            )
        self.button1 = go('button1')
        self.button1.connect(
            "clicked",
            self.button_1_clicked,
            )
        self.button2 = go('button2')
        self.button2.connect(
            "clicked",
            self.button_2_clicked,
            )
        self.button3 = go('button3')
        self.button3.connect(
            "clicked",
            self.button_3_clicked,
            )
        self.button4 = go('button4')
        self.button4.connect(
            "clicked",
            self.button_4_clicked,
            )
        self.button5 = go('button5')
        self.button5.connect(
            "clicked",
            self.button_5_clicked,
            )
        self.label_number_of_records = go('labelNumberOfRecords')
        self.buttonPasteWithHeader = go('buttonPasteWithHeader')
        self.buttonPasteWithHeader.connect(
            "clicked",
            self.button_paste_with_header_clicked,
            "cool button 2"
            )
        self.button_print_selected_labels = go('buttonPrintSelectedLabels')
        self.button_print_selected_labels.connect(
            "clicked",
            self.print_selected_labels,
            )
        self.button_print_every_label = go('buttonPrintEveryLabel')
        self.button_print_every_label.connect(
            "clicked",
            self.print_every_label,
            )

        self.button_sheet_selected_labels = go('buttonSheetSelectedLabels')
        self.button_sheet_selected_labels.connect(
            "clicked",
            self.sheet_selected_labels,
            )
        self.button_sheet_every_label = go('buttonSheetEveryLabel')
        self.button_sheet_every_label.connect(
            "clicked",
            self.sheet_every_label,
            )

        self.cb_printers = go('comboboxPrinters')
        self.cb_models = go('comboboxModels')
        self.cb_models.connect(
            "changed",
            self.cb_models_selection_changed,
            "dummy"
            )

        self.treeview_params = go('treeviewParams')
        self.treeview_params.get_selection().connect(
            "changed",
            self.params_tree_selection_changed,
            "dummy"
            )
        self.treeview_params.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            dnd_targets,
            Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE
            )
        self.treeview_params.connect('drag_data_get', on_drag_data_get)

        fill_combo(self.cb_printers, self.printers.keys())
        fill_combo(self.cb_models, self.model_names)
        self.set_date_code_to_today()
        # Connect signals
        self.builder.connect_signals(self)
        self.cb_models_selection_changed(self.cb_models)
        # Configure interface
        self.window.connect('delete-event', lambda x, y: Gtk.main_quit())
        # GObject.timeout_add(1000, self.update_mac_pool)
        # Everything is ready
        self.window.show()
        self.password_generator = PasswordGenerator(src1, src2)

    def get_copies(self):
        try:
            return int(self.entry_copies.get_text())
        except:
            return 1

    def print_selected_labels(self, widget, data=None):
        self.print_labels(self.get_selected_labels(), self.get_copies())

    def sheet_selected_labels(self, widget, data=None):
        self.sheet_labels(self.get_selected_labels(), self.get_copies())

    def get_selected_labels(self):
        # print("print_selected_label")
        selection = self.treeview_data_table.get_selection()
        # print('selection:', selection)
        (model, pathlist) = selection.get_selected_rows()
        labels = []
        for path in pathlist:
            # print('path:', path)
            tree_iter = model.get_iter(path)
            # print('tree_iter:', tree_iter)
            label_data = {}
            for param in self.get_parameters():
                value = None
                for index, column in enumerate(
                        self.treeview_data_table.get_columns()):
                    if column.get_title() == param:
                        value = model.get_value(
                            tree_iter,
                            index
                            )
                        break
                # print('value:', value)
                if value is None:
                    if param == 'date code':
                        entry_date_code_text = self.entry_date_code.get_text()
                        if len(entry_date_code_text) == 4:
                            value = entry_date_code_text
                if value:
                    label_data[param] = value
                else:
                    print(repr(param), 'parameter is missing')
                    return
            labels.append(label_data)
        return labels

    def get_label_from_row(self, row):
        label_data = {}
        for param in self.get_parameters():
            value = None
            for index, column in enumerate(
                    self.treeview_data_table.get_columns()):
                if column.get_title() == param:
                    value = row[index]
                    break
            # print('value:', value)
            if value:
                label_data[param] = value
            else:
                print(repr(param), 'parameter is missing')
                return
        return label_data

    def print_every_label(self, widget, data=None):
        labels = []
        for row in self.treeview_data_table.get_model():
            label_data = self.get_label_from_row(row)
            if label_data:
                labels.append(label_data)
        self.print_labels(labels, self.get_copies())

    def sheet_every_label(self, widget, data=None):
        labels = []
        for row in self.treeview_data_table.get_model():
            label_data = self.get_label_from_row(row)
            if label_data:
                labels.append(label_data)
        self.sheet_labels(labels, self.get_copies())

    def button_1_clicked(self, widget, data=None):
        self.entry_copies.set_text('1')

    def button_2_clicked(self, widget, data=None):
        self.entry_copies.set_text('2')

    def button_3_clicked(self, widget, data=None):
        self.entry_copies.set_text('3')

    def button_4_clicked(self, widget, data=None):
        self.entry_copies.set_text('4')

    def button_5_clicked(self, widget, data=None):
        self.entry_copies.set_text('5')

    def get_number_of_records(self):
        return len(self.treeview_data_table.get_model())

    def adjust_number_records(self):
        number_of_records = self.get_number_of_records()
        self.label_number_of_records.set_text(str(number_of_records))
        self.button_print_every_label.set_sensitive(number_of_records > 0)
        self.button_sheet_every_label.set_sensitive(number_of_records > 0)

    def button_paste_clicked(self, widget, data=None):
        # print("button_paste_clicked - %s" % data)
        set_treeview_from_clipboard(self.treeview_data_table, with_header=False)
        self.adjust_number_records()

    def button_paste_with_header_clicked(self, widget, data=None):
        # print("button_paste_with_header_clicked - %s" % data)
        set_treeview_from_clipboard(self.treeview_data_table, with_header=True)
        self.adjust_number_records()

    def data_tree_selection_changed(self, widget, data=None):
        # print("selection_changed - %s" % data)
        csr = widget.count_selected_rows()
        # print("count_selected_rows:", csr)
        self.button_print_selected_labels.set_sensitive(csr > 0)
        self.button_sheet_selected_labels.set_sensitive(csr > 0)
        self.update_label_preview()

    def update_label_preview(self):
        selected_labels = self.get_selected_labels()
        if selected_labels is not None and len(selected_labels) == 1:
            size_x, size_y = self.get_label_size()

            self.selected_label = selected_labels[0]
            # print('preview_label:', self.selected_label)
            self.selected_label_file = self.get_label_file(self.selected_label)
            # Gnm.textview_set_text(
            #     self.textview.set_text,
            #     self.selected_label_file
            #     )
            self.textview.get_buffer().set_text(self.selected_label_file)
            # self.preview_label_file(self.get_label_file(label))
            self.drawingarea.set_size_request(
                self.dpi * size_x / 25.4,
                self.dpi * size_y / 25.4
                )
            self.drawingarea.queue_draw()
            # self.window.reshow_with_initial_size()
        else:
            self.textview.get_buffer().set_text('')
            self.drawingarea.set_size_request(-1, -1)
            self.selected_label = None
            self.label_file = None
            # self.window.reshow_with_initial_size()

    # def preview_label_file(self, label_file):
    #     print('preview_label_file', label_file)
    #     for line in label_file.splitlines():
    #         if line.startswith('T'):
    #             print ("preview", line)
    #             control, text = line.split(';', 1)
    #             control = control[2:]
    #             pos_x, pos_y, rotate, font, size = control.split(',')
    #             print ("text", pos_x, pos_y, rotate, font, size, text)

    def on_draw(self, wid, cr):
        # frame it!
        size_x, size_y = self.get_label_size()
        cr.set_line_width(1)
        cr.rectangle(
            0,
            0,
            self.dpi * size_x / 25.4,
            self.dpi * size_y / 25.4,
            )
        cr.set_source_rgb(255, 255, 255)
        cr.fill()
        # cr.set_source_rgb(0, 0, 0)
        # cr.stroke()
        self.draw_label_file(
            (
                self.selected_label_file
                if hasattr(self, 'selected_label_file') else ''
                ),
            cr,
            self.dpi
            )

    def draw_label_file(self, label_file, cr, dpi):
        # cr.set_font_size(dpi * 6.0 / 72)
        unit = default_unit
        for line in label_file.splitlines():
            if line.startswith('m'):
                unit = line[2:]  # m or i
                if unit == 'i':
                    # print('unit is inch')
                    pass
                elif unit == 'm':
                    # print('unit is mm')
                    pass
                else:
                    print('unsupported line:', line)
            elif line.startswith('J'):
                if len(line) != 1:
                    print('unsupported line:', line)
            elif line.startswith('S'):
                # size TBD
                pass
            elif line.startswith('O'):
                # TBD
                pass
            elif line.startswith('A'):
                # TBD
                pass
            elif line.startswith('T'):
                # print("preview", line)
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate, font, size = control.split(',')
                # print("text", pos_x, pos_y, rotate, font, size, text)
                cr.move_to(dots(pos_x, unit, dpi), dots(pos_y, unit, dpi))
                size = dots(size, unit, dpi)
                cr.set_font_size(size)
                if font == '3':
                    cr.select_font_face(
                        "Swiss 721",
                        cairo.FONT_SLANT_NORMAL,
                        cairo.FONT_WEIGHT_NORMAL,
                        )
                elif font == '4':
                    cr.select_font_face(
                        "Monospace 821",
                        cairo.FONT_SLANT_NORMAL,
                        cairo.FONT_WEIGHT_NORMAL,
                        )
                elif font == '5':
                    cr.select_font_face(
                        "Swiss 721",
                        cairo.FONT_SLANT_NORMAL,
                        cairo.FONT_WEIGHT_BOLD,
                        )
                cr.set_source_rgb(0, 0, 0)
                cr.show_text(text)
            elif line.startswith('G'):
                # print("preview", line)
                # G 0.05,0.64,0;R:1.9,0.33,0.0066666
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate = control.split(',')
                function, function_params = text.split(':', 1)
                if function == 'R':
                    width, height, thickness = function_params.split(',', 2)
                    cr.set_line_width(dots(thickness, unit, dpi))
                    cr.rectangle(
                        dots(pos_x, unit, dpi),
                        dots(pos_y, unit, dpi),
                        dots(width, unit, dpi),
                        dots(height, unit, dpi),
                        )
                    cr.set_source_rgb(0, 0, 0)
                    cr.stroke()
                else:
                    print('unsupported line:', line)
            elif line.startswith('B'):
                # print("preview", line)
                # B 1.39,0.08,0,QRCODE+EL{error level}+MODEL2,.0133333;{qrcode}
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate, barcode, size = control.split(',', 4)
                barcode_opts = barcode.split('+')
                if barcode_opts[0] == 'QRCODE':
                    qr_model = 1
                    error_correction = 'L'
                    for opt in barcode_opts:
                        if opt == 'QRCODE':
                            pass
                        elif opt == 'MODEL1':
                            qr_model = 1
                        elif opt == 'MODEL2':
                            qr_model = 2
                        elif opt == 'ELL':
                            error_correction = 'L'
                        elif opt == 'ELM':
                            error_correction = 'M'
                        elif opt == 'ELQ':
                            error_correction = 'Q'
                        elif opt == 'ELH':
                            error_correction = 'H'
                        else:
                            print('unsupported qrcode option:', opt)
                    box_size = int(0.5 + dots(size, unit, dpi))
                    if qr_model != 2:
                        print('unsupported QR Code model:', qr_model)
                    try:
                        with tempfile.NamedTemporaryFile(
                            suffix='.png',
                            # suffix='.svg',
                            prefix='qr',
                            delete=False
                                ) as qr_file:
                            qr_file_name = qr_file.name
                        args = [
                            'qrencode',
                            '-o', qr_file_name,
                            '-l', error_correction,
                            '-m', '0',
                            # '-t', 'svg',
                            '-s', str(box_size),
                            # '"' + text + '"'
                            text
                            ]
                        # print("args:", args)
                        subprocess.check_call(args)
                        surface = cairo.ImageSurface.create_from_png(
                            qr_file_name
                            )
                    except:
                        traceback.print_exc(file=sys.stdout)
                    finally:
                        try:
                            if os.path.isfile(qr_file_name):
                                os.remove(qr_file_name)
                                # print('removed temporary file:', qr_file_name)
                        except:
                            print('cannot remove temporary file:', qr_file_name)
                            traceback.print_exc(file=sys.stdout)
                    # cr.set_antialias(cairo.ANTIALIAS_NONE)
                    cr.set_source_surface(
                        surface,
                        dots(pos_x, unit, dpi),
                        dots(pos_y, unit, dpi)
                        )
                    cr.paint()
                    # print('cr', dir(cr))
                else:
                    print('unsupported line:', line)
            else:
                print('unsupported line:', line)

    def params_tree_selection_changed(self, widget, data=None):
        # print("selection_changed - %s" % data)
        # csr = widget.count_selected_rows()
        # print("params_tree_selection_changed count_selected_rows:", csr)
        pass

    def get_label_name(self):
        # print('get_label_name')
        for model in self.models:
            # print('model:', model)
            for model_name, label_name in iteritems(model):
                # print('model name:', model_name, 'label name:', label_name)
                if model_name == self.model:
                    # print('label_name:', label_name)
                    return label_name

    def get_label(self):
        # print('get_label')
        label = self.labels[self.get_label_name()]
        # print('label:', label)
        return label

    def get_parameters(self):
        # print('get_parameters')
        try:
            parameters = self.get_label()['parameters']
            # print('parameters:', parameters)
            return parameters
        except:
            # silent fail, because it is possible to have a label without param
            return []

    def get_label_type(self):
        return self.label_types[self.get_label()['label type']]

    def get_label_size(self):
        label_type = self.get_label_type()
        return label_type['width'], label_type['height']

    def cb_models_selection_changed(self, widget, data=None):
        # print("cb_models_selection_changed - %s" % data)
        tree_iter = widget.get_active_iter()
        if tree_iter is not None:
            self.model = widget.get_model()[tree_iter][0]
            # print("Selected model:", self.model)
            set_treeview_from_texts(
                self.treeview_params,
                self.get_parameters()
                )
        self.update_label_preview()

    def button_today_clicked(self, widget, data=None):
        # print("button_today_clicked - %s" % data)
        self.set_date_code_to_today()

    def set_date_code_to_today(self):
        today = datetime.date.today()
        todays_week = today.strftime("%W")
        todays_year = today.strftime("%Y")
        todays_date_code = str(todays_year[-2:]) + str(todays_week)
        self.entry_date_code.set_text(todays_date_code)

    def get_label_file(self, label_params, x_shift=None, y_shift=None):
        for param in self.get_parameters():
            if param not in label_params:
                raise ValueError('label parameter %s is missing' % repr(param))

        label_format_params = {'model': self.model}
        label_name = self.get_label_name()
        if label_name == "#02 with meid":
            label_format_params['qrcode'] = (
                "#02:" +
                self.model + ":" +
                label_params['date code'] + ":" +
                label_params['ethernet mac address'] + ":" +
                label_params['snap mac address'] + ":" +
                label_params['wifi mac address'] + ":" +
                label_params['modem meid']
                )
            label_format_params['error level'] = 'H'
        elif label_name == "#02 without meid":
            label_format_params['qrcode'] = (
                "#02:" +
                self.model + ":" +
                label_params['date code'] + ":" +
                label_params['ethernet mac address'] + ":" +
                label_params['snap mac address'] + ":" +
                label_params['wifi mac address']
                )
            label_format_params['error level'] = 'H'
        elif label_name == "#08":
            label_format_params['qrcode'] = (
                "#08:" +
                self.model + ":" +
                label_params['date code'] + ":" +
                label_params['ethernet mac address last 6'][-6:] + ":" +
                label_params['snap node address'][-6:] + ":" +
                label_params['wifi mac address'] + ":" +
                label_params['modem meid'] + ":" +
                label_params['sim iccid'] + ":"
                )
            label_format_params['error level'] = (
                'Q' if len(label_format_params['qrcode']) >= 80 else 'H'
                )
        elif label_name == "#05":
            label_format_params['error level'] = 'L'
        label_format_params.update(label_params)
        # print('label_format_params:', label_format_params)
        return shift_and_format_items(
            self.get_label()['items'],
            x_shift=x_shift,
            y_shift=y_shift,
            label_format_params=label_format_params
            )

    def print_label_file_to(self, label_file, printer_config):
        # create random file name
        filename = ''.join(random.choice(string.digits) for _ in range(9))
        # print("random file name:", filename)
        if 'ftp' in printer_config:
            ftp_config = printer_config['ftp']
            ftp_connection = None
            try:
                ftp_connection = ftplib.FTP(
                    ftp_config['ip address']
                    )
                ftp_connection.login(
                    ftp_config['user name'],
                    ftp_config['password']
                    )
                ftp_connection.cwd("execute")
                ftp_connection.storlines(
                    'STOR ' + filename,
                    StringIO(label_file)
                    )
            except:
                print("FTP Error")
                traceback.print_exc(file=sys.stdout)
            finally:
                if ftp_connection is not None:
                    ftp_connection.close()

    def get_selected_printer_config(self):
        tree_iter = self.cb_printers.get_active_iter()
        if tree_iter is not None:
            selected_printer_name = self.cb_printers.get_model()[tree_iter][0]
            return self.printers[selected_printer_name]

    def print_label_file(self, label_file):
        # print('print_label_file:\n"""\n%s"""' % label_file)
        selected_printer_config = self.get_selected_printer_config()
        if selected_printer_config is not None:
            self.print_label_file_to(label_file, selected_printer_config)

    def print_labels(self, labels, copies=1):
        # print('printing model:', self.model, 'labels:', labels)
        x_shift = 0
        lpr = self.get_label_type()['lpr']
        total_width = self.get_label_type()['width']
        if lpr > 1:
            total_width += (lpr - 1) * self.get_label_type()['hrepeat']
        # print('lpr:', lpr)
        label_file = None
        if labels is not None:
            for label in labels:
                for copy in range(copies):
                    current_label_file = self.get_label_file(
                        label,
                        (x_shift * self.get_label_type()['hrepeat'])
                        if x_shift else None
                        )
                    if label_file is None:
                        label_file = 'm m\n'
                        label_file += 'J\n'
                        label_file += (
                            'S l1;0,0,' +
                            str(self.get_label_type()['height']) + ',' +
                            str(self.get_label_type()['vrepeat']) + ',' +
                            str(total_width) +
                            '\n'
                            )
                        label_file += 'O R\n'
                    label_file += current_label_file
                    x_shift += 1
                    if x_shift >= lpr:
                        label_file += 'A 1\n'
                        self.print_label_file(label_file)
                        label_file = None
                        x_shift = 0
        if label_file is not None:
            label_file += 'A 1\n'
            self.print_label_file(label_file)
            label_file = None
            x_shift = 0

    def button_generate_clicked(self, widget, data=None):
        # print("button_generate_clicked - %s" % data)
        set_treeview_from_text_data(
            self.treeview_data_table,
            self.generate_records(),
            with_header=True,
            )
        self.adjust_number_records()

    def entry_records_changed(self, widget, data=None):
        self.button_generate.set_sensitive(False)
        records = int(self.entry_records.get_text())
        # print('records', records)
        self.button_generate.set_sensitive(records > 0)

    def generate_records(self):
        try:
            records = int(self.entry_records.get_text())
        except ValueError:
            records = 0
        # print('records', records)
        parameters = self.get_parameters()
        retval = '\t'.join(parameters) + '\n'
        # reset the generators
        self.incremented_params = {}
        self.fixed_params = {}
        for record in range(records):
            more = record < (records - 1)
            retval += '\t'.join(
                [
                    self.generate_parameter(
                        parameter,
                        more
                        ) for parameter in parameters
                    ]
                ) + '\n'
        return retval

    def generate_parameter(self, parameter, more):
        if parameter == 'date code':
            if parameter in self.fixed_params:
                new_param = self.fixed_params[parameter]
            else:
                while True:
                    new_param, fixed = gtk_input(
                        self.window,
                        parameter,
                        default=self.entry_date_code.get_text(),
                        fixed_option=True,
                        )
                    if new_param is None:
                        return None
                    new_param = new_param.strip().upper()
                    # TBD check if it is a scanned qr code
                    try:
                        check_parameter(parameter, new_param)
                        break
                    except ValueError as error:
                        print(error)
                if fixed:
                    self.fixed_params[parameter] = new_param
            return new_param
        elif parameter in ['user password', 'wifi password']:
            return self.password_generator.generate_password()
        elif parameter in [
            'ethernet mac address',
            'snap mac address',
            'wifi mac address',
            'snap node address',
            'modem meid',
                ]:
            if parameter in self.incremented_params:
                new_param = hex_increment(self.incremented_params[parameter])
                self.incremented_params[parameter] = new_param
            else:
                while True:
                    new_param, auto_inc = gtk_input(
                        self.window,
                        parameter,
                        auto_inc_option=more,
                        )
                    if new_param is None:
                        return None
                    new_param = new_param.strip().upper()
                    # TBD check if it is a scanned qr code
                    try:
                        check_parameter(parameter, new_param)
                        break
                    except ValueError as error:
                        print(error)
                if auto_inc:
                    self.incremented_params[parameter] = new_param
            return new_param
        elif parameter in [
            'sim iccid',
                ]:
            if parameter in self.incremented_params:
                new_param = dec_increment(self.incremented_params[parameter])
                self.incremented_params[parameter] = new_param
            else:
                while True:
                    new_param, auto_inc = gtk_input(
                        self.window,
                        parameter,
                        auto_inc_option=more,
                        )
                    if new_param is None:
                        return None
                    new_param = new_param.strip().upper()
                    # TBD check if it is a scanned qr code
                    try:
                        check_parameter(parameter, new_param)
                        break
                    except ValueError as error:
                        print(error)
                if auto_inc:
                    self.incremented_params[parameter] = new_param
            return new_param
        else:
            # TBD pop-up info dialog instead
            raise ValueError(
                "cannot generate parameter: %s" %
                repr(parameter)
                )

    def sheet_labels_old(self, labels, copies=1):
        self.operation = Gtk.PrintOperation()
        print_data = {
            'num_pages': 0,
            'labels': labels,
            'copies': copies
            }

        # self.operation.set_use_full_page(False)
        # self.operation.set_unit(Gtk.Unit.POINTS)
        # self.operation.set_embed_page_setup(True)

        settings = Gtk.PrintSettings()

        doc_dir = GLib.get_user_special_dir(
            GLib.UserDirectory.DIRECTORY_DOCUMENTS
            )
        if doc_dir is None:
            doc_dir = GLib.get_home_dir()
        if settings.get(Gtk.PRINT_SETTINGS_OUTPUT_FILE_FORMAT) == 'ps':
            ext = '.ps'
        elif settings.get(Gtk.PRINT_SETTINGS_OUTPUT_FILE_FORMAT) == 'svg':
            ext = '.svg'
        else:
            ext = '.pdf'

        uri = "file://%s/gtk-demo%s" % (doc_dir, ext)
        settings.set(Gtk.PRINT_SETTINGS_OUTPUT_URI, uri)
        settings.set_resolution(self.dpi)
        # settings.set_printer_lpi(150.0)
        self.operation.set_print_settings(settings)
        print(
            'res x and y:',
            self.operation.get_print_settings().get_resolution_x(),
            self.operation.get_print_settings().get_resolution_y()
            )

        self.operation.connect('begin-print', self.begin_print, print_data)
        self.operation.connect('draw-page', self.draw_page, print_data)
        self.operation.connect('end-print', self.end_print, print_data)

        result = self.operation.run(
            Gtk.PrintOperationAction.PRINT_DIALOG,
            self.window
            )
        if result == Gtk.PrintOperationResult.ERROR:
            message = self.operation.get_error()
            dialog = Gtk.MessageDialog(self.window,
                                       0,
                                       Gtk.MessageType.ERROR,
                                       Gtk.ButtonsType.CLOSE,
                                       message)
            dialog.run()
            dialog.destroy()

    def begin_print(self, operation, print_ctx, print_data):
        page_dpi_y = print_ctx.get_dpi_y()
        page_dpi_x = print_ctx.get_dpi_x()
        print('page_dpi_x:', page_dpi_x, 'page_dpi_y:', page_dpi_y)

        page_y_px = print_ctx.get_height(
            ) - self.HEADER_HEIGHT - self.HEADER_GAP
        page_x_px = print_ctx.get_width()
        print('page_x_px:', page_x_px, 'page_y_px:', page_y_px)

        page_dpi_y = print_ctx.get_dpi_y()
        page_dpi_x = print_ctx.get_dpi_x()
        print('page_dpi_x:', page_dpi_x, 'page_dpi_y:', page_dpi_y)

        page_x_mm = 25.4 * page_x_px / page_dpi_x
        page_y_mm = 25.4 * page_y_px / page_dpi_y
        print('page_x_mm:', page_x_mm, 'page_y_mm:', page_y_mm)

        label_x_mm, label_y_mm = self.get_label_size()
        print('label_x_mm:', label_x_mm, 'label_y_mm:', label_y_mm)

        labels_x = int(page_x_mm / label_x_mm)
        labels_y = int(page_y_mm / label_y_mm)
        print_data['labels_page'] = labels_x * labels_y
        print(
            'labels_x:', labels_x,
            'labels_y:', labels_y,
            'labels_page:', print_data['labels_page']
            )
        print_data['n_labels'] = len(
            print_data['labels']
            ) * print_data['copies']
        print_data['num_pages'] = (
            print_data['n_labels'] - 1
            ) / print_data['labels_page'] + 1
        print('num_pages:', print_data['num_pages'])
        operation.set_n_pages(print_data['num_pages'])

    def draw_page(self, operation, print_ctx, page_num, print_data):
        dpi_x = print_ctx.get_dpi_x()
        print(
            "draw_page dpi_x:", dpi_x,
            operation.get_print_settings().get_resolution_x()
            # cairo_surface_get_fallback_resolution
            )
        cr = print_ctx.get_cairo_context()
        width = print_ctx.get_width()
        cr.rectangle(0, 0, width, self.HEADER_HEIGHT)
        cr.set_source_rgb(0.8, 0.8, 0.8)
        cr.fill_preserve()
        cr.set_source_rgb(0, 0, 0)
        cr.set_line_width(1)
        cr.stroke()
        layout = print_ctx.create_pango_layout()
        desc = Pango.FontDescription('sans 14')
        layout.set_font_description(desc)
        # layout.set_text(print_data['filename'], -1)
        (text_width, text_height) = layout.get_pixel_size()
        if text_width > width:
            layout.set_width(width)
            layout.set_ellipsize(Pango.EllipsizeMode.START)
            (text_width, text_height) = layout.get_pixel_size()
        cr.move_to((width - text_width) / 2,
                   (self.HEADER_HEIGHT - text_height) / 2)
        PangoCairo.show_layout(cr, layout)
        page_str = "%d/%d" % (page_num + 1, print_data['num_pages'])
        layout.set_text(page_str, -1)
        layout.set_width(-1)
        (text_width, text_height) = layout.get_pixel_size()
        cr.move_to(width - text_width - 4,
                   (self.HEADER_HEIGHT - text_height) / 2)
        PangoCairo.show_layout(cr, layout)
        layout = print_ctx.create_pango_layout()
        # desc = Pango.FontDescription('monospace')
        # desc.set_size(print_data['font_size'] * Pango.SCALE)
        # layout.set_font_description(desc)
        cr.move_to(0, self.HEADER_HEIGHT + self.HEADER_GAP)
        for label in print_data['labels']:
            label_file = self.get_label_file(label)
            print('label_file:', label_file)
            self.draw_label_file(label_file, cr, dpi_x)
            return
        # lines_pp = int(print_data['lines_per_page'])
        # num_lines = print_data['num_lines']
        # data_lines = print_data['lines']
        # font_size = print_data['font_size']
        # line = page_num * lines_pp
        # for i in range(lines_pp):
        #     if line >= num_lines:
        #         break
        #     layout.set_text(data_lines[line], -1)
        #     PangoCairo.show_layout(cr, layout)
        #     cr.rel_move_to(0, font_size)
        #     line += 1

    def end_print(self, operation, print_ctx, print_data):
        pass

    def sheet_labels(self, labels, copies=1):
        # typical paper and margin sizes
        # letter size: 8.5 x 11 in.
        # printable area on letter: 8.2 x 10.6 in.
        # top margin .22 in.
        # bottom margin: .18 in
        # sides margin .15 in
        paper_x_mm = 8.5 * 25.4
        paper_y_mm = 11 * 25.4
        side_margin_mm = 0.15 * 25.4
        top_margin_mm = 0.22 * 25.4
        bottom_margin_mm = 0.15 * 25.4
        printable_x_mm = paper_x_mm - side_margin_mm - side_margin_mm
        printable_y_mm = paper_y_mm - top_margin_mm - bottom_margin_mm
        label_x_mm, label_y_mm = self.get_label_size()
        labels_x = int(printable_x_mm / label_x_mm)
        labels_y = int(printable_y_mm / label_y_mm)
        # labels_page = labels_x * labels_y
        doc_dir = GLib.get_user_special_dir(
            GLib.UserDirectory.DIRECTORY_DOCUMENTS
            )
        if doc_dir is None:
            doc_dir = GLib.get_home_dir()
        pdf_file = os.path.join(doc_dir, 'gtk-demo.pdf')
        # print('pdf_file:', pdf_file)
        pdfdpi = 600.0
        ps = cairo.PDFSurface(
            pdf_file,
            pdfdpi * paper_x_mm / 25.4,
            pdfdpi * paper_y_mm / 25.4
            )
        cr = cairo.Context(ps)
        x_pos = 0
        y_pos = 0
        for label in labels:
            for copy in range(copies):
                x_shift_mm = side_margin_mm + x_pos * label_x_mm
                y_shift_mm = top_margin_mm + y_pos * label_y_mm
                label_file = self.get_label_file(
                    label,
                    x_shift=x_shift_mm,
                    y_shift=y_shift_mm
                    )
                # frame it!
                cr.set_line_width(1)
                cr.rectangle(
                    pdfdpi * x_shift_mm / 25.4,
                    pdfdpi * y_shift_mm / 25.4,
                    pdfdpi * label_x_mm / 25.4,
                    pdfdpi * label_y_mm / 25.4,
                    )
                # cr.set_source_rgb(255, 255, 255)
                # cr.fill()
                cr.set_source_rgb(0, 0, 0)
                cr.stroke()

                self.draw_label_file(label_file, cr, pdfdpi)
                x_pos += 1
                if x_pos >= labels_x:
                    x_pos = 0
                    y_pos += 1
                    if y_pos >= labels_y:
                        y_pos = 0
                        cr.show_page()
        if x_pos > 0 or y_pos > 0:
            cr.show_page()
        ps.finish()
        if os.path.isfile(pdf_file):
            start_file(pdf_file)


def synapse_labels():
    basicConfig(level=INFO)
    gui = LabelsGui()
    assert gui
    Gtk.main()


if __name__ == "__main__":
    synapse_labels()
